package com.example.cpspowerprobe;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.charset.Charset;
import java.util.HashMap;
import java.util.Iterator;

import android.annotation.SuppressLint;
import android.app.PendingIntent;
import android.content.Context;
import android.content.Intent;
import android.hardware.usb.UsbDevice;
import android.hardware.usb.UsbDeviceConnection;
import android.hardware.usb.UsbEndpoint;
import android.hardware.usb.UsbInterface;
import android.hardware.usb.UsbManager;
import android.hardware.usb.UsbRequest;
import android.util.Log;
import android.widget.Toast;


public class Probe {

	///////////////////////  Constants /////////////////////////////
	
	private static final String ACTION_USB_PERMISSION =
		    "com.android.example.USB_PERMISSION";
	
	// Probe Status
	public static final int PROBE_STS_NOTCONNECTED =	0;
	public static final int PROBE_STS_INPROGRESS =		1;
	public static final int PROBE_STS_CONNECTED =		2;
	
	// Commands
	public static final byte GET_PROPERIES_CMD =	 0x20;
	public static final byte GET_PARAM_CMD = 		 0x21;
	public static final byte GET_FIRMWARE_INFO_CMD = 0x22;
	public static final byte IDENT_DEVICE_CMD = 	 (byte)0x80;
	public static final byte GET_STATUS_CMD =		 (byte)0x81;
	public static final byte GET_VALUES_CMD = 		 (byte)0x82;
	public static final byte READ_EEPROM_CMD = 		 (byte)0x83;
	public static final byte WRITE_EEPROM_CMD =		 (byte)0x84;
	
	
        public static final int EEPROM_SIZE = (60+1);
	public static final int USB_BUFFER_SIZE = 64;
	
	public static final int NVALS = 				5;
	public static final double MAXVALIND = 			1e10;
	public static final double MINVALIND = 			-1e10;
		
	public static final int MAXPARAMS =				3;
	public static final int T_PARAMS_INDEX =		0;
	public static final int DC_PARAMS_INDEX =   	1;
	public static final int AC_PARAMS_INDEX =		2;
		
	// Param types
	public static final byte NONE =					0;
	public static final byte VOLTAGE =				1;
	public static final byte CURRENT =				2;
	public static final byte TEMPERATURE =			3;
		
	private static final int INT_FIELD_SIZE = 		4;
							
	///////////////////////  Properties /////////////////////////////
	
	private UsbDevice mProbe = null;	
	private UsbDeviceConnection mConnection = null;
	private UsbEndpoint mEpWrite = null;
	private UsbEndpoint mEpRead = null;
	private UsbInterface mInterface = null;	
	private CPS_device_properties mDeviceProperties = null;
	private CPS_param_value[] mParamValues = null;	
	private CPS_calibration_buffer mCalBuf = null;
	private CPS_probe_data mProbeData = null;	
	private CPS_probe_data mCorrData = null;
	private CPS_probe_data mZeroData = null;
	private CPS_probe_data mFilterPrev = null;
	private int mStatus = PROBE_STS_NOTCONNECTED;	
	private String mFWVer;
		
	//////////////////////  Sub-Classes /////////////////////////////
	
	public class CPS_probe_data {
		public CPS_probe_data(CPS_probe_data src) {
			t = 	src.t;
			dc = 	src.dc;
			ac = 	src.ac;
			min = 	src.min;
			max = 	src.max;
		}
		public CPS_probe_data() {}
		public double
			t,
			dc,
			ac,
			min,
			max;
	}		
	
	public class CPS_calibration_buffer {
		public float
			dc_scale,
			dc_offset,
			dc_scale_t,
			dc_offset_t,
			ac_scale,
			ac_offset,
			ac_scale_t,
			ac_offset_t,
		    dc_scale2,
		    ac_scale2,
		    dc_scale3, //unused
		    ac_scale3, //unused
		    reserved1,
		    reserved2,
		    reserved3;
		public CPS_calibration_buffer(ByteBuffer buf) {
			if (buf == null) {
				dc_scale = 0.0f;
				ac_scale = 0.0f;
				dc_offset = 0.0f;
				dc_scale_t = 0.0f;
				dc_offset_t = 0.0f;
				ac_scale_t = 0.0f;
				ac_offset_t = 0.0f;
				dc_scale2 = 0;
				ac_scale2 = 0;
				dc_scale3 = 0;
				ac_scale3 = 0;
				reserved1 = 0;
				reserved2 = 0;
				reserved3 = 0;
			} else{
				buf.order(ByteOrder.LITTLE_ENDIAN);
				buf.position(1); // skip echoed command
				dc_scale = buf.getFloat();
				dc_offset = buf.getFloat();
				dc_scale_t = buf.getFloat();
				dc_offset_t = buf.getFloat();
				ac_scale = buf.getFloat();
				ac_offset = buf.getFloat();
				ac_scale_t = buf.getFloat();
				ac_offset_t = buf.getFloat();
				dc_scale2 = buf.getFloat();
				ac_scale2 = buf.getFloat();
			}
		}
		public boolean checkValues() {
			if (dc_scale < -2.0 || 
				dc_scale > 2.0 ||
				ac_scale < -2.0 ||
				ac_scale > 2.0) {
				return false;
			}
			return true;
		}
		public void setDefault() {
			dc_scale = 
					ac_scale = 0;
			dc_offset =
					dc_scale_t =
					dc_offset_t =
					ac_offset =
					ac_scale_t =
					ac_offset_t = 0;
			ac_scale2 = 0;
			dc_scale2 = 0;
			ac_scale3 = 0;
			dc_scale3 = 0;
		}
	}				
	
	public class CPS_device_properties {
		public int model;		// model number
		public byte type;		// type, one of device_types
		public byte version;	// device version
		public byte nparams;	// number of params
		public CPS_device_properties(ByteBuffer buf) {			
			buf.order(ByteOrder.LITTLE_ENDIAN);
			buf.position(1); // skip echoed command
			model = buf.getInt();
			type = buf.get();
			version = buf.get();
			nparams = buf.get();			
		}
	}
	
	public class CPS_param_value {
		public byte
			type,               // type, one of param_types
        	dir,                // directions, one of param_direction
        	unit,               // user's unit, one of param_units
        	bits,               // number of bits
        	sigval;             // value is signed 
		public int
			max_binary,         // binary value corresponding to maximum user's value
			zero_binary,        // binary value corresponding to zero user's value
			min_binary,         // binary value corresponding to minimum user's value
			over_binary,        // overflow code
			under_binary,       // underflow code
			max_user,           // maximum value in user's units
			zero_user,          // zero value in user's units
			min_user;           // minimum value in user's units
		public CPS_param_value(ByteBuffer buf) {			
			buf.order(ByteOrder.LITTLE_ENDIAN);
			buf.position(1); // skip echoed command
			type = buf.get();
			dir = buf.get();
			unit = buf.get();
			bits = buf.get();
			sigval = buf.get();
			max_binary = buf.getInt();
			zero_binary = buf.getInt();
			min_binary = buf.getInt();
			over_binary = buf.getInt();
			under_binary = buf.getInt();
			max_user = buf.getInt();
			zero_user = buf.getInt();
			min_user = buf.getInt();								
		}
	}
	
	
	////////////////////// Public API /////////////////////////////
	
	// Constructor
	public Probe() {		
		mProbeData = new CPS_probe_data();       
		mCorrData = new CPS_probe_data();
		mFilterPrev = new CPS_probe_data();
        mParamValues = new CPS_param_value[MAXPARAMS];              
	}
	
	// Disconnect from USB probe
	public void Close() {
		if (mStatus == PROBE_STS_CONNECTED) {
			mConnection.releaseInterface(mInterface);
			mConnection.close();
		}		
		mConnection = null;
		mInterface = null;
		mEpRead = null;
		mEpWrite = null;
		mProbe = null;
		mStatus = PROBE_STS_NOTCONNECTED;
	}
	
	public int GetConnStatus() {
		return mStatus;
	}
	
	public int getMaxUserDC(){		
		return mParamValues[DC_PARAMS_INDEX].max_user;		
	}
	
	public int getMaxUserAC(){
		return mParamValues[AC_PARAMS_INDEX].max_user;
	}
	
	// Scan for USB probe
	public boolean FindProbe(Intent intent, Context context, UsbManager usbManager) {    
		
		HashMap<String, UsbDevice> deviceList = usbManager.getDeviceList();
		Iterator<UsbDevice> deviceIterator = deviceList.values().iterator();
		while(deviceIterator.hasNext()){
		    UsbDevice device = deviceIterator.next();
		    if (device.getVendorId() == 0x1943 && device.getProductId() == 0x9001) {		    				    				    
		    	// probe found - request permission
		    	mStatus = PROBE_STS_INPROGRESS;
		    	usbManager.requestPermission(device, PendingIntent.getBroadcast(context, 0, new Intent(ACTION_USB_PERMISSION), Intent.FLAG_ACTIVITY_NEW_TASK));			    	
		    	return true;
		    }
		}
		
		// no probe found
		return false;				
	 }   
	
	// connect to USB probe
	public boolean UsbConnect(UsbDevice device, Context context, UsbManager usbManager) {
		
		mProbe = device;
				
		if (!usbManager.hasPermission(mProbe)){					    	
	    	return false;
	    } 
			    	    	 	    	  
	    mInterface = mProbe.getInterface(0);
	    if (mInterface == null) {
	    	return false;
	    }

	    mEpRead = mInterface.getEndpoint(0);
	    if (mEpRead == null) {
	    	return false;
	    }

	    mEpWrite = mInterface.getEndpoint(1);
	    if (mEpWrite == null) {
	    	return false;
	    }

	    mConnection = usbManager.openDevice(mProbe);
	    if (mConnection == null) {
	    	return false;
	    }	 
	    	    	   
	    try {
	    	mConnection.claimInterface(mInterface, true);
		} catch (Exception e) {
	    	Toast.makeText(context, (CharSequence) "Probe Exception", Toast.LENGTH_SHORT).show();
	    	return false;
	    }	    	    

	    mCalBuf = new CPS_calibration_buffer( ReadCalibration() );	    	
	    	   	    	    	   
    	if (!mCalBuf.checkValues()) {
    		Toast.makeText(context, (CharSequence) "Invalid calibration values", Toast.LENGTH_SHORT).show();
    		mCalBuf.setDefault();
    		if (WriteCalibration()) {
    			Toast.makeText(context, (CharSequence) "Defaults applied", Toast.LENGTH_SHORT).show();
    		}
    	}
    		   
    	getProperties();
        getParams();
        getFirmware();
	            
        mStatus = PROBE_STS_CONNECTED;

        return true;
	}	
	
	public CPS_calibration_buffer GetCalibration() {
		return mCalBuf;
	}
	
	public void SetCalibration(CPS_calibration_buffer buf) {
		mCalBuf.ac_offset = buf.ac_offset;
		mCalBuf.ac_scale = buf.ac_scale;
		mCalBuf.dc_offset = buf.dc_offset;
		mCalBuf.dc_scale = buf.dc_scale;
		mCalBuf.ac_scale2 = buf.ac_scale2;
		mCalBuf.dc_scale2 = buf.dc_scale2;
		WriteCalibration();
	}

	public CPS_probe_data GetRawValues() {
		getProbeData();
		return mProbeData;
	}
	
	// Get corrected data
	public CPS_probe_data GetValues(double timebase) {		
		getProbeData();
		filterVals(timebase);
		
		CorrectValues();
		if (mZeroData != null)
			zeroValues();				
		return mCorrData;
	}
	
	public int GetStatus(){								
		return (int) UsbRead(GET_STATUS_CMD).get(1);			
	}
	
	public String GetFWVer() {
		return mFWVer;
	}
		
	@SuppressLint("DefaultLocale")
	public String GetDeviceInfo() {
		
		return String.format("%04d:%04x", mDeviceProperties.model, mDeviceProperties.version);
	}
	
	public void SetZeros()
	{
		getProbeData();
		CorrectValues();
				
		mZeroData = new CPS_probe_data(mCorrData);
				
		mZeroData.t = 0;		
	}
	
	public void ClearZeros()
	{		
		mZeroData = null;
	}
		
	////////////////////// Internal Use Methods /////////////////////////////	
	
	// filter "vals[]" with "timebase" time constant
	private void filterVals(double timebase)
	{	  
	  double dv;	  

	  dv = (mProbeData.dc - mFilterPrev.dc) / timebase;
	  mFilterPrev.dc = mProbeData.dc = mFilterPrev.dc + dv;
	  
	  dv = (mProbeData.ac - mFilterPrev.ac) / timebase;
	  mFilterPrev.ac = mProbeData.ac = mFilterPrev.ac + dv;
	  
	  dv = (mProbeData.min - mFilterPrev.min) / timebase;
	  mFilterPrev.min = mProbeData.min = mFilterPrev.min + dv;
	  
	  dv = (mProbeData.max - mFilterPrev.max) / timebase;
	  mFilterPrev.max = mProbeData.max = mFilterPrev.max + dv;
	  
	  dv = (mProbeData.t - mFilterPrev.t) / timebase;
	  mFilterPrev.t = mProbeData.t = mFilterPrev.t + dv;
	  
	}
	
	private void zeroValues()
	{		 											
		mCorrData.dc -= 	mZeroData.dc;
		mCorrData.ac -= 	mZeroData.ac;
		mCorrData.min -= 	mZeroData.min;
		mCorrData.max -= 	mZeroData.max;
	}		
		
	private void getFirmware() {
		ByteBuffer fwData;		
		fwData = UsbRead(GET_FIRMWARE_INFO_CMD);
				
		mFWVer = new String( fwData.array(), Charset.forName("UTF-8") );
		mFWVer = mFWVer.substring(1, mFWVer.indexOf(0));
		
	}
	
	private void getProbeData() {
		int v;
		ByteBuffer data = UsbRead(GET_VALUES_CMD);
		
		data.position(0);
		
		data.order(ByteOrder.LITTLE_ENDIAN);		 	
		
		data.get(); // skip command
		v = data.getInt();
		mProbeData.t = bintouser(v, T_PARAMS_INDEX);	
		v = data.getInt();
		mProbeData.dc = bintouser(v, DC_PARAMS_INDEX);
		v = data.getInt();
		mProbeData.ac = bintouser(v, AC_PARAMS_INDEX);
		v = data.getInt();
		mProbeData.min = bintouser(v, DC_PARAMS_INDEX);
		v = data.getInt();
		mProbeData.max = bintouser(v, DC_PARAMS_INDEX);			
	}
	    
	 private void getProperties() {
		 ByteBuffer props = UsbRead(GET_PROPERIES_CMD);
    	    	
		 mDeviceProperties = new CPS_device_properties(props);

		 if ( mDeviceProperties.nparams > MAXPARAMS) {
    		mDeviceProperties.nparams = MAXPARAMS;
		 }    	
	 }
	    
	 private void getParams() {    	
		 for (byte cnt = 0; cnt < MAXPARAMS; cnt++) {    		
			 ByteBuffer param = UsbRead(GET_PARAM_CMD, cnt);    		
			 mParamValues[cnt] = new CPS_param_value(param);    		
		 } 
		 
		 
		 //Log.i("MyActivity", String.format("min_b: %d max_b: %d", mParamValues[DC_PARAMS_INDEX].min_binary, mParamValues[DC_PARAMS_INDEX].max_binary));
	 }
	    
	 private ByteBuffer ReadCalibration() {
	    	// read eeprom
	    	ByteBuffer eeData;
	    	int att, cnt;
	    	byte sum;    	
	    	
	    	for (att = 0; att < 5; att++) {
	    		Log.d("read calib", "att" + att);
	    		eeData = UsbRead(READ_EEPROM_CMD);	    			    		
	    		eeData = UsbRead(READ_EEPROM_CMD);	    		
	    		if (eeData != null) {    			
	    			// compute checksum
	    			for (sum = 0, cnt = 1; cnt < EEPROM_SIZE + 1; cnt++ ) {
	    				sum += eeData.get(cnt);
	    			}    			
	    			if (sum == 0) {    			
	    				return eeData;
	    			}
	    		}    		
	    	} // for
	    	Log.d("read calib", "failed!");
	    	return null;
	    }
	    
	 private boolean WriteCalibration() {
		 ByteBuffer data = ByteBuffer.allocate(USB_BUFFER_SIZE);
		 int cnt, sum;

		 data.order(ByteOrder.LITTLE_ENDIAN);
		 data.put(WRITE_EEPROM_CMD);
		 data.putFloat(mCalBuf.dc_scale);    	
		 data.putFloat(mCalBuf.dc_offset);
		 data.putFloat(mCalBuf.dc_scale_t);
		 data.putFloat(mCalBuf.dc_offset_t);
		 data.putFloat(mCalBuf.ac_scale);
		 data.putFloat(mCalBuf.ac_offset);
		 data.putFloat(mCalBuf.ac_scale_t);
		 data.putFloat(mCalBuf.ac_offset_t);
		 data.putFloat(mCalBuf.dc_scale2); //
		 data.putFloat(mCalBuf.ac_scale2); //mCalBuf.ac_scale2
		 data.putFloat(0); //mCalBuf.dc_scale3
		 data.putFloat(0); //mCalBuf.ac_scale3
		 data.putFloat(0); //mCalBuf.reserve1
		 data.putFloat(0); //mCalBuf.reserve2
		 data.putFloat(0); //mCalBuf.reserve3

		 for (sum = 0, cnt = 1; cnt < EEPROM_SIZE; cnt++){
			 sum += data.get(cnt);
		 }
		 data.put((byte)-sum);
		 UsbWrite(data);
		 return true;
	 }
	    
	 // convert variable length value to integer
		private int bintoint(int bin, int bits, int sigval)
		{			
		  bin <<= (INT_FIELD_SIZE * 8) - bits;
		  if(sigval != 0 )
		    return bin >> (INT_FIELD_SIZE * 8) - bits; // signed
		  return bin >>> (INT_FIELD_SIZE * 8) - bits;  // unsigned
	    	    	
		}
		
		private double bintouser(int bin, int parnum) {			
			double scale, offset, result;			
			
			if (mParamValues[parnum].type == NONE)
				return 0;
			if (bin == mParamValues[parnum].over_binary)
				return mParamValues[parnum].max_user + MAXVALIND;
			if ( bin == mParamValues[parnum].under_binary)
				return mParamValues[parnum].min_user + MINVALIND;			
			
			bin = bintoint(bin, mParamValues[parnum].bits, mParamValues[parnum].sigval);						

			scale = (double)(mParamValues[parnum].max_user - mParamValues[parnum].min_user) /
			        (mParamValues[parnum].max_binary - mParamValues[parnum].min_binary);
			offset = mParamValues[parnum].max_user - mParamValues[parnum].max_binary * scale;
			
			result = bin * scale;
			result += offset;
			
			bin = (int)result;
			
			//Log.i("MyActivity", String.format("result %f", result));						
			
			if(bin < mParamValues[parnum].min_user)
				return mParamValues[parnum].min_user + MINVALIND;
			if(bin > mParamValues[parnum].max_user)
			    return mParamValues[parnum].max_user + MAXVALIND;

			return result;
		}
						
		private void CorrectValues() {
			if (mCorrData.dc > 0) {
				mCorrData.dc = 	(mProbeData.dc * (mCalBuf.dc_scale + 1)) + (mProbeData.dc * mProbeData.dc * mCalBuf.dc_scale2) + mCalBuf.dc_offset;
			} else {
				mCorrData.dc = 	(mProbeData.dc * (mCalBuf.dc_scale + 1)) - (mProbeData.dc * mProbeData.dc * mCalBuf.dc_scale2) + mCalBuf.dc_offset;
			}
			if (mCorrData.ac > 0) {
				mCorrData.ac = 	(mProbeData.ac * (mCalBuf.ac_scale + 1)) +
						(mProbeData.ac * mProbeData.ac * mCalBuf.ac_scale2) + mCalBuf.ac_offset;
			} else {
				mCorrData.ac = 	(mProbeData.ac * (mCalBuf.ac_scale + 1)) -
						(mProbeData.ac * mProbeData.ac * mCalBuf.ac_scale2) + mCalBuf.ac_offset;
			}
			mCorrData.min = mProbeData.min * mCalBuf.dc_scale + mCalBuf.dc_offset;
			mCorrData.max = mProbeData.max * mCalBuf.dc_scale + mCalBuf.dc_offset;		
			mCorrData.t = 	mProbeData.t * 		1 + 				0;  // no correction for temp											
		}
				
		private boolean UsbCmd(byte cmd, byte idx) {
			byte[] bytes = new byte[USB_BUFFER_SIZE];
			bytes[0] = cmd;
			bytes[1] = idx;
			return (mConnection.bulkTransfer(mEpWrite, bytes, bytes.length, 0) == USB_BUFFER_SIZE);	
		}		
		
		private boolean UsbWrite(ByteBuffer data) {			
			return (mConnection.bulkTransfer(mEpWrite, data.array(), data.capacity(), 0) == USB_BUFFER_SIZE);
		}				
		
		private ByteBuffer UsbRead(byte cmd, byte idx){					
			
			ByteBuffer buffer = ByteBuffer.allocate(USB_BUFFER_SIZE);
			UsbRequest request = new UsbRequest();
			request.initialize(mConnection, mEpRead);
		    //queue a request on the interrupt
			buffer.clear();
			request.queue(buffer, USB_BUFFER_SIZE);						
			// send command
			UsbCmd(cmd, idx);
			
			// wait for reply
			if (mConnection.requestWait() == request) {
				return buffer;
			} else {
				request.close();
				Log.d("usbread", "request wait fail");
				return null;
			}
				
		}
		private ByteBuffer UsbRead(byte cmd){ return UsbRead(cmd, (byte)0); } // overload for 1 byte read commands	    


}
